require ("price_list_generator")
--require ("tile_data")

script.on_init(function (event)

  game.map_settings.pollution.enabled = false
  local surface = game.surfaces[1]
  surface.always_day = true
  global.sell_rate = 0.93
  global.buy_rate = 1.07
  global.price_per_energy = 1/200000
  global.bounding_limit = 60 --Don't let the player go out further than 60x square
  global.average_period = 10*60
  global.buy_chests = {}
  global.sell_chests = {}
  global.accumulators = {}
  global.income = {}
  global.expenses = {}
  global.average_expense = {}
  global.average_income = {}
  generate_price_list()
  --save_map_data(100,50,50)
  --recreate_map(tiles,entities)
  set_research()
  for k, v in pairs(surface.find_entities()) do
    v.destructible = false
    v.minable = false
    v.rotatable = false
    if v.name == "logistic-chest-requester" then
      table.insert(global.buy_chests, v)
    end
    if v.name == "logistic-chest-passive-provider" then
      table.insert(global.sell_chests, v)
    end
    if v.name =="electric-energy-interface" then
    
    v.power_production = 0
    v.power_usage = 0
    v.electric_buffer_size = 5*10^9
    v.energy = 2.55*10^9
    v.electric_output_flow_limit = 50000000
    v.electric_input_flow_limit = 50000000
    
      table.insert(global.accumulators, v)
    end
  end
  global.cash = {}
  for k, force in pairs (game.forces) do
    force.manual_crafting_speed_modifier = -1 --Disable handcrafting 
    global.cash[force.name] = 10000
  end

end)

script.on_configuration_changed(generate_price_list)

script.on_event(defines.events.on_player_created, function (event)
  
  local player = game.players[event.player_index]
  local character = player.character
  player.character = nil
  if character then 
    character.destroy()
  end
  
  if (#game.players <= 1) then
    game.show_message_dialog{text = {"factory-intro"}}
  end
  

end)

script.on_event(defines.events.on_tick, function(event)
  index = (game.tick%global.average_period)+1
  reset_income(index)
  reset_expenses(index)
  electric_exchange(global.accumulators[(game.tick % #global.accumulators)+1])
  sell_items(global.sell_chests[(game.tick % #global.sell_chests)+1])
  buy_items(global.buy_chests[(game.tick % #global.buy_chests)+1])
  calculate_income_and_expenses(index)
  if game.tick % 3 == 0 then
    for k, player in pairs (game.players) do
      update_cash(player)
    end
  end
  check_bounding(global.bounding_limit)
end)

function sell_items(chest)
  local inventory = chest.get_inventory(defines.inventory.chest)
  for k, contents in pairs (inventory.get_contents()) do
    if global.price_list[k] then
      local cost = round((contents*global.price_list[k])*global.sell_rate)
      chest.remove_item{name = k, count = contents}
      global.cash[chest.force.name] = global.cash[chest.force.name] + cost
      update_income(cost, chest.force)
    end
  end
end

function buy_items(chest)
local inventory = chest.get_inventory(defines.inventory.chest)
  for k = 1,10 do
    local stack = chest.get_request_slot(k)
    if stack then
      if inventory.get_item_count(stack.name) < stack.count then 
      local buy_count = stack.count - inventory.get_item_count(stack.name)
      if global.price_list[stack.name] then
        local price = global.price_list[stack.name]
        local cost = round(buy_count*price*global.buy_rate)
        if global.cash[chest.force.name] > cost and inventory.can_insert(stack) then
          global.cash[chest.force.name] = global.cash[chest.force.name] - cost
          chest.insert{name = stack.name, count = buy_count}
          update_expenses(cost, chest.force)
        end
      else
      --chest.insert{name = stack.name, count = buy_count}
      end
      
      end
    end
  end
end

function electric_exchange(accumulator)
  local difference = accumulator.energy - 2.55*10^9
  local cost = difference*global.price_per_energy
  if cost > 0 then
    global.cash[accumulator.force.name] = global.cash[accumulator.force.name] + (cost*global.sell_rate)
    accumulator.energy = 2.55*10^9
    update_income(cost*global.sell_rate, accumulator.force)
  elseif cost < 0 then
    global.cash[accumulator.force.name] = global.cash[accumulator.force.name] + (cost*global.buy_rate)
    accumulator.energy = 2.55*10^9
    update_expenses(cost*global.buy_rate, accumulator.force)
  end
end

function update_cash(player)
  if player.gui.top.cash == nil then
    local frame = player.gui.top.add{name = "cash", type = "frame", direction = "vertical"}
    frame.style.minimal_width = 150
    frame.style.left_padding = 10
    frame.style.top_padding = 8
    frame.add
      {
        name="cash_amount",
        type = "label",
        caption={"", {"cash"}, " ", comma_value(round(global.cash[player.force.name])) }
      } 
    frame.add
      {
        name="income_amount",
        type = "label",
        caption={"", {"cash"}, " ", comma_value(round((global.average_income[player.force.name]/global.average_period)*60)),"/m" }
      } 
    frame.add
      {
        name="expense_amount",
        type = "label",
        caption={"", {"cash"}, " ", comma_value(round((global.average_expense[player.force.name]/global.average_period)*60)),"/m" }
      }
    frame.add
      {
        name="profit_amount",
        type = "label",
        caption={"", {"profit"}, " ", comma_value(round(((global.average_income[player.force.name]/global.average_period)*60) - (global.average_expense[player.force.name]/global.average_period)*60)),"/m" }
      }      
    frame.add
      {
        name = "toggle_price_list",
        type = "button",
        caption = {"toggle-price-list"}
      }
  else
    player.gui.top.cash.cash_amount.caption = {"", {"cash"}, " ", comma_value(round(global.cash[player.force.name])) }
    player.gui.top.cash.income_amount.caption = {"", {"income"}, " ", comma_value(round((global.average_income[player.force.name]/global.average_period)*60*60)),"/m" }    player.gui.top.cash.expense_amount.caption = {"", {"expense"}, " ", comma_value(round((global.average_expense[player.force.name]/global.average_period)*60*60)),"/m" }player.gui.top.cash.profit_amount.caption = {"", {"profit"}, " ", comma_value(round(((global.average_income[player.force.name]/global.average_period)*60*60) - (global.average_expense[player.force.name]/global.average_period)*60*60)),"/m" }
  end
end
script.on_event(defines.events.on_gui_checked_state_changed, function(event)
local gui = event.element
local player = game.players[event.player_index]
  if gui.name == "purchase_amount_5_check" then
    player.gui.left.price_list.purchase_amount_stack_check.state = false
  elseif gui.name == "purchase_amount_stack_check" then
    player.gui.left.price_list.purchase_amount_5_check.state = false
  end
end)

script.on_event(defines.events.on_gui_click, function(event)
local gui = event.element
local player = game.players[event.player_index]
  if gui.name == "query_price" then
    query_price(player)
  elseif global.price_list[gui.name] then
    if player.gui.left.price_list.purchase_amount_5_check.state then
      player_purchase(player,gui.name,5)
    elseif player.gui.left.price_list.purchase_amount_stack_check.state then
      player_purchase(player,gui.name,game.item_prototypes[gui.name].stack_size)
    else 
      player_purchase(player,gui.name,1)
    end
  elseif gui.name == "toggle_price_list" then
    generate_price_table(player)
  end
  
end)

function comma_value(amount)
  local formatted = amount
  while true do  
    formatted, k = string.gsub(formatted, "^(-?%d+)(%d%d%d)", '%1,%2')
    if (k==0) then
      break
    end
  end
  return formatted
end

function check_bounding(limit)
  for k, player in pairs(game.players) do
    if player.position.x > limit then 
      player.teleport{limit, player.position.y}
    end
    if player.position.x < -limit then 
      player.teleport{-limit, player.position.y}
    end
    if player.position.y > limit then 
      player.teleport{player.position.x, limit}
    end
    if player.position.y < -limit then 
      player.teleport{player.position.x, -limit}
    end
  end
end



function save_map_data(distance, save_offset_x, save_offset_y)

--This exports current map data as an array of entities and tiles

local data = "tiles = \n{\n"
  for X = -distance,distance do
    for Y = -distance, distance do
      local tile = game.surfaces[1].get_tile(X,Y)
      local name = tile.name
      if name ~= "out-of-map" then
      local position = tile.position
      data = data.."  {name = \""..name.."\", position = {"..position.x+save_offset_x..","..position.y+save_offset_y.."}},\n"
      end
    end
  end
  data = data.."\n}\n\n".."entities = \n{\n"
  for k, entity in pairs (game.surfaces[1].find_entities({{-distance,-distance},{distance,distance}})) do
    local name = entity.name
    local position = entity.position
    local direction = entity.direction
    local force = entity.force
    if entity.name == "express-loader" then
      local loader_type = entity.loader_type
      data = data.."  {name = \""..name.."\", position = {"..position.x+save_offset_x..","..position.y+save_offset_y.."}, force = \""..force.name.."\", direction = "..direction..", type = \""..loader_type.."\"},\n"
    else
    data = data.."  {name = \""..name.."\", position = {"..position.x+save_offset_x..","..position.y+save_offset_y.."}, force = \""..force.name.."\", direction = "..direction.."},\n"
    end
  end
  data = data.."\n}"
  game.write_file("tile_data.lua",data)
end



function recreate_map(tiles,entities)

--This creates a section of map using an array of tiles and entities

  game.surfaces[1].set_tiles(tiles,true)
  for k, entity in pairs (entities) do
    local v = game.surfaces[1].create_entity(entity)
    v.destructible = false
    v.minable = false
    v.rotatable = false
    if v.name == "logistic-chest-requester" then
      table.insert(global.buy_chests, v)
    end
    if v.name == "logistic-chest-passive-provider" then
      table.insert(global.sell_chests, v)
    end
    if v.name =="electric-energy-interface" then
    
    v.power_production = 0
    v.power_usage = 0
    v.electric_buffer_size = 5*10^9
    v.energy = 2.55*10^9
    v.electric_output_flow_limit = 50000000
    v.electric_input_flow_limit = 50000000
    
      table.insert(global.accumulators, v)
    end
  end

end

function set_research()

--[=[
Because we don't want research like shooting speed or damage upgrades,
we use this to disable everything and only enabled certain types of research
]=]--

  for i, force in pairs (game.forces) do
    for k, research in pairs (force.technologies) do
      research.enabled = false
    end
  end
  
  for k, research in pairs (game.forces.player.technologies) do
    local unlock_this = false
    for j, effect in pairs (research.effects) do
      if effect.type == "unlock-recipe"
      or effect.type == "inserter-stack-size-bonus"
      or effect.type == "stack-inserter-size-bonus"
      or effect.type == "worker-robot-storage"
      or effect.type == "worker-robot-speed" then
        unlock_this = true
      end
    end
    if unlock_this then
      for i, force in pairs (game.forces) do
        force.technologies[research.name].enabled = true
      end
      unlock_prerequisite(research)
    end
  end
end

function unlock_prerequisite(research)
--This enables the prerequisite technologies for the given research
  for k, prerequisite in pairs (research.prerequisites) do
    for i, force in pairs (game.forces) do
      force.technologies[prerequisite.name].enabled = true
    end
  unlock_prerequisite(prerequisite) --Loops back to enabled the prerequisite of the prerequisite
  end
end


function round(n)
  local v = 0.01*math.floor((n*100)+0.5)
  return v
end

function generate_price_table(player)
  if not player.gui.left.price_list then
    local frame = player.gui.left.add{type = "frame", name = "price_list", caption = {"",{"price-list"},""}, direction = "vertical"}
    frame.style.maximal_height = 600
       frame.add{type = "label", name = "purchase_amount_5", caption = {"",{"buy-5"},""}}
    frame.add{type = "checkbox", name = "purchase_amount_5_check", state = false}
    frame.add{type = "label", name = "purchase_amount_stack", caption = {"",{"buy-stack"},""}}
    frame.add{type = "checkbox", name = "purchase_amount_stack_check", state = false}
    local header_table = frame.add{type = "table", colspan = 4, name = "header_table"}
    header_table.style.cell_spacing = 16
    header_table.style.left_padding = 0
    header_table.style.bottom_padding = 0
    

    
    header_table.add{type = "label", name = "item_icon", caption = {"", {"buy"}, ""}}
    header_table.add{type = "label", name = "item_name", caption = {"", {"item-name"},""}}
    header_table.add{type = "label", name = "buy_price", caption = {"","    ", {"buy-price"},""}}
    header_table.add{type = "label", name = "sell_price", caption = {"","    ", {"sell-price"},""}}
    local scroll_pane = frame.add{type = "scroll-pane", name = "price_scrollpane", vertical_scroll_policy = "auto"}
    scroll_pane.style.maximal_height = 450
    local price_table = scroll_pane.add{type = "table", colspan = 4, name = "price_table"}
    price_table.style.cell_spacing = 1
    --price_table.left_padding = 5

    for name, price in spairs(global.price_list) do
      if game.item_prototypes[name] then
        local icon = price_table.add{type = "sprite-button", name = name, sprite = "item/"..name}
        icon.style.minimal_width = 40
        icon.style.minimal_height = 40
        price_table.add{type = "label", name = name.."name", caption = game.item_prototypes[name].localised_name}
        price_table.add{type = "label", name = name.."buy_price", caption = comma_value(round(global.price_list[name]*global.buy_rate))}
        price_table.add{type = "label", name = name.."sell_price", caption = comma_value(round(global.price_list[name]*global.sell_rate))}
      end
    end
  else
    player.gui.left.price_list.destroy()
  end
end

function spairs(t, order)
    -- collect the keys
    local keys = {}
    for k in pairs(t) do keys[#keys+1] = k end

    -- if order function given, sort by it by passing the table and keys a, b,
    -- otherwise just sort the keys 
    if order then
        table.sort(keys, function(a,b) return order(t, a, b) end)
    else
        table.sort(keys)
    end

    -- return the iterator function
    local i = 0
    return function()
        i = i + 1
        if keys[i] then
            return keys[i], t[keys[i]]
        end
    end
end

function player_purchase(player,name,amount)
  local price = global.price_list[name]
  local cost = round(price*global.buy_rate*amount)
  if global.cash[player.force.name] > cost and player.can_insert({name = name, count = amount}) then
    global.cash[player.force.name] = global.cash[player.force.name] - cost
    player.insert{name = name, count = amount}
  end
end

function reset_income(index)
  for k, force in pairs (game.forces) do
    if global.income[force.name] == nil then 
      global.income[force.name] = {} 
      for i = 1, global.average_period do
        global.income[force.name][i] = 0
      end
    end
    if not global.average_income[force.name] then global.average_income[force.name] = 0 end
    global.average_income[force.name] = global.average_income[force.name] - global.income[force.name][index]
    global.income[force.name][index] = 0
  end
end

function reset_expenses(index)
  for k, force in pairs (game.forces) do
    if not global.expenses[force.name] then 
      global.expenses[force.name] = {} 
      for i = 1, global.average_period do
        global.expenses[force.name][i] = 0
      end
    end
    if not global.average_expense[force.name] then global.average_expense[force.name] = 0 end
    global.average_expense[force.name] = global.average_expense[force.name] - global.expenses[force.name][index]
    global.expenses[force.name][index] = 0
  end
end

function update_income(cost, force)
  global.income[force.name][index] = global.income[force.name][index] + cost
end

function update_expenses(cost, force)
    global.expenses[force.name][index] = global.expenses[force.name][index] + cost
end

function calculate_income_and_expenses(index)
  for k, force in pairs (game.forces) do
    global.average_expense[force.name] = global.average_expense[force.name] + global.expenses[force.name][index]
    global.average_income[force.name] = global.average_income[force.name] + global.income[force.name][index]
  end
end






